"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const graphql_1 = require("graphql");
const fs_extra_1 = require("fs-extra");
const chokidar_1 = require("chokidar");
const glob = __importStar(require("glob"));
const ast_1 = require("graphql-tool-utilities/ast");
const print_1 = require("./print");
class Builder extends events_1.EventEmitter {
    constructor(_a) {
        var { graphQLFiles, schemaPath } = _a, options = __rest(_a, ["graphQLFiles", "schemaPath"]);
        super();
        this.watching = false;
        this.documentCache = new Map();
        this.globs = graphQLFiles;
        this.schemaPath = schemaPath;
        this.options = options;
    }
    once(event, handler) {
        return super.once(event, handler);
    }
    on(event, handler) {
        return super.on(event, handler);
    }
    emit(event, ...args) {
        return super.emit(event, ...args);
    }
    run({ watch: watchGlobs = false } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            this.watching = watchGlobs;
            const { globs } = this;
            const update = (file) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield this.updateDocumentForFile(file);
                }
                catch (error) {
                    this.emit('error', error);
                    return;
                }
                yield this.generateDocumentTypes();
            });
            if (watchGlobs) {
                const documentWatcher = chokidar_1.watch(globs);
                documentWatcher.on('ready', () => {
                    documentWatcher.on('add', update);
                    documentWatcher.on('change', update);
                    documentWatcher.on('unlink', (file) => __awaiter(this, void 0, void 0, function* () {
                        this.removeDocumentForFile(file);
                        yield this.generateDocumentTypes();
                    }));
                });
                const schemaWatcher = chokidar_1.watch(this.schemaPath);
                schemaWatcher.on('ready', () => {
                    schemaWatcher.on('change', () => __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield this.updateSchema();
                            yield this.generateSchemaTypes();
                            yield this.generateDocumentTypes();
                        }
                        catch (error) {
                            // intentional noop
                        }
                    }));
                });
            }
            try {
                yield this.updateSchema();
                yield this.generateSchemaTypes();
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            try {
                yield Promise.all(glob.sync(globs).map(this.updateDocumentForFile.bind(this)));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            yield this.generateDocumentTypes();
        });
    }
    generateSchemaTypes() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('schema:start');
            const definition = print_1.printSchema(this.schema);
            yield fs_extra_1.writeFile(this.options.schemaTypesPath, definition);
            this.emit('schema:end');
        });
    }
    generateDocumentTypes() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('start');
            let ast;
            try {
                ast = ast_1.compile(this.schema, graphql_1.concatAST(Array.from(this.documentCache.values())));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            const fileMap = groupOperationsAndFragmentsByFile(ast);
            try {
                const buildResults = yield Promise.all(Object.keys(fileMap).map((key) => __awaiter(this, void 0, void 0, function* () {
                    const file = fileMap[key];
                    let definition;
                    try {
                        definition = print_1.printDocument(file, ast, this.options);
                    }
                    catch ({ message }) {
                        const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
                        this.emit('error', error);
                        throw error;
                    }
                    const definitionPath = `${file.path}.d.ts`;
                    yield fs_extra_1.writeFile(definitionPath, definition);
                    return {
                        documentPath: file.path,
                        definitionPath,
                        operation: file.operation,
                        fragments: file.fragments,
                    };
                })));
                for (const buildResult of buildResults) {
                    this.emit('build', buildResult);
                }
            }
            catch (error) {
                return;
            }
            this.emit('end');
        });
    }
    updateSchema() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const schemaJSON = yield fs_extra_1.readJSON(this.schemaPath);
                this.schema = graphql_1.buildClientSchema(schemaJSON.data);
            }
            catch (error) {
                const parseError = new Error(`Error parsing '${this.schemaPath}':\n\n${error.message.replace(/Syntax Error GraphQL \(.*?\) /, '')}`);
                throw parseError;
            }
        });
    }
    updateDocumentForFile(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const contents = yield fs_extra_1.readFile(file, 'utf8');
            if (contents.trim().length === 0) {
                return;
            }
            const document = graphql_1.parse(new graphql_1.Source(contents, file));
            this.documentCache.set(file, document);
        });
    }
    removeDocumentForFile(file) {
        this.documentCache.delete(file);
    }
}
exports.Builder = Builder;
function groupOperationsAndFragmentsByFile({ operations, fragments, }) {
    const map = {};
    Object.keys(operations).forEach((name) => {
        const operation = operations[name];
        const file = map[operation.filePath] || {
            path: operation.filePath,
            operation,
            fragments: [],
        };
        map[operation.filePath] = file;
    });
    Object.keys(fragments).forEach((name) => {
        const fragment = fragments[name];
        const file = map[fragment.filePath] || {
            path: fragment.filePath,
            operation: undefined,
            fragments: [],
        };
        file.fragments.push(fragment);
        map[fragment.filePath] = file;
    });
    return map;
}
